//Original Code:
//Ralf Habel 2007
//Implementation of the paper "Instant Animated Grass" published in WSCG Journal 2007.
//modified to fit in engine

float4x4 matVP;
float2 BlockOffset;
float3 CamPos;
float Time; //Runs from 0 to 10 per second.

float PatchScale;
float2 PatchOffset;
float FieldSize;
float BlockSize;
float2 GrassScale;
float WindScale;
float4 BlendMask;

float3 LightDir;
float4 AmbientColor;
float4 LightColor;

int TexFilter;
int TexAnisotropy;

texture HeightMap;
sampler SHeightMap = sampler_state
{
    Texture = <HeightMap>;
    AddressU = Clamp;
    AddressV = Clamp;
    MipFilter = NONE;
    MagFilter = LINEAR;
    MinFilter = LINEAR;
};

texture NormalMap;
sampler SNormalMap = sampler_state
{
    Texture = <NormalMap>;
    AddressU = Clamp;
    AddressV = Clamp;
    MipFilter = NONE;
    MagFilter = LINEAR;
    MinFilter = LINEAR;
};

texture Blades;
sampler SBlades = sampler_state
{
    Texture = <Blades>;
    AddressU = Wrap;
    AddressV = Wrap;
    MipFilter = LINEAR;
    MagFilter = POINT;
  //  MinFilter = (TexFilter);
    MinFilter = POINT;
  //  MaxAnisotropy = (TexAnisotropy);
};

texture WindNoise;
sampler SWindNoise = sampler_state
{
    Texture = <WindNoise>;
    AddressU = Wrap;
    AddressV = Wrap;
    MipFilter = NONE;
    MagFilter = LINEAR;
    MinFilter = LINEAR;
};

texture Blend;
sampler SBlend = sampler_state
{
    Texture = <Blend>;
    AddressU = Clamp;
    AddressV = Clamp;
    MipFilter = LINEAR;
    MagFilter = LINEAR;
    MinFilter = (TexFilter);
    MaxAnisotropy = (TexAnisotropy);
};

//Set everything which is constant in the fragment shader, the values used are for the moderate height data set.

#define MAX_RAYDEPTH 8 //Number of iterations.

#define PLANE_NUM 16.0 //Number of grass slice grid planes per unit in tangent space.

#define PLANE_NUM_INV (1.0/PLANE_NUM)
#define PLANE_NUM_INV_DIV2 (PLANE_NUM_INV/2)

#define GRASS_SLICE_NUM 8 // Number of grass slices in texture grassblades.

#define GRASS_SLICE_NUM_INV (1.0/GRASS_SLICE_NUM)
#define GRASS_SLICE_NUM_INV_DIV2 (GRASS_SLICE_NUM_INV/2)

#define GRASSDEPTH GRASS_SLICE_NUM_INV //Depth set to inverse of number of grass slices so no stretching occurs.

#define TC1_TO_TC2_RATIO 8 //Ratio of texture coordinate set 1 to texture coordinate set 2, used for the animation lookup.

#define PREMULT (GRASS_SLICE_NUM_INV*PLANE_NUM) //Saves a multiply in the shader.

#define AVERAGE_COLOR float4(0.32156,0.513725,0.0941176,1.0) //Used to fill remaining opacity, can be replaced by a texture lookup.

struct VS_IN
{
    float4 Pos : POSITION;
};

struct VS_OUT
{
    float4 Position     : POSITION;
    float4 Diffuse      : COLOR0;
    float2 TexCoordGrass : TEXCOORD0;
    float2 TexCoordWind  : TEXCOORD1;
    float2 TexCoordBlend  : TEXCOORD2;
    float3 EyeDirTan    : TEXCOORD3;
    float4 PositionViewProj : TEXCOORD4;
};

VS_OUT BillboardGrass_VP(VS_IN i, uniform bool bShadow)
{
    VS_OUT o;

    float2 p = i.Pos * PatchScale + PatchOffset;
    float2 hmtc = (p + 1.5) / (BlockSize+3);
    float2 nmtc = (p + 0.5) / (BlockSize+1);
    o.TexCoordBlend = nmtc;
    float2 BlockPos = p / BlockSize;
    o.TexCoordWind = BlockPos * WindScale + Time;
    o.TexCoordGrass = BlockPos * GrassScale;
    p = p * FieldSize + BlockOffset;

    float3 wpos = float3(p, tex2Dlod(SHeightMap, float4(hmtc, 0, 0)).r);
    wpos.z += GRASSDEPTH * BlockSize *FieldSize / GrassScale;
    o.Position = mul(float4(wpos, 1), matVP);

    float2 nxy = tex2Dlod(SNormalMap, float4(nmtc, 0, 0)) * 2 - 1;
    float3 Normal = float3(nxy, 1 - nxy.x*nxy.x - nxy.y*nxy.y);
    //float3 Normal = float3(0, 0, 1);
    float3 BiNormal = float3(0, -1, 0);
    float3 Tangent = cross(Normal, BiNormal);
    //float3 Tangent = float3(1, 0, 0);
    BiNormal = cross(Normal, Tangent);

    o.PositionViewProj = o.Position;

//    o.TexCoord = texCoord;
//    o.TexCoord2 = float2((texCoord2.x+time*0.2)/2,(texCoord2.y+time*0.2)/2); // offset second texture coordinate
                                        // according to time for wind texture
    float3 eyeDirO = wpos - CamPos; //eye vector in object space
    
    float3x3 TBNMatrix = float3x3(Tangent, BiNormal, Normal); 

    o.EyeDirTan = normalize(mul(TBNMatrix, eyeDirO)); // eye vector in tangent space

    o.Diffuse = saturate(dot(Normal, -LightDir)); 

    return o;
}


//---------------------------------------------------------------------------------------------------------------------

struct PS_OUT
{
    float4 Color : COLOR;
    float Depth : DEPTH;
};

PS_OUT BillboardGrass_PS(VS_OUT i, uniform bool bShadow)
{   
    PS_OUT o;

    //Initialize increments/decrements and per fragment constants
    float4 color = float4(0, 0, 0, 0);

    float2 plane_offset = float2(0, 0);                 
    float3 rayEntry = float3(i.TexCoordGrass, 0);
    float zOffset = 0.0;
    bool zFlag = 1;

    if(i.EyeDirTan.z > 0)
    {
        o.Depth = i.PositionViewProj.z/i.PositionViewProj.w;
        o.Color = float4(1,1,0,0);
        //return o;
    }
    else
    {

    //The signs of eyeDirTan determines if we increment or decrement along the tangent space axis
    //plane_correct, planemod and pre_dir_correct are used to avoid unneccessary if-conditions. 
    
    float2 sign = float2(sign(i.EyeDirTan.x),sign(i.EyeDirTan.y));  
    float2 plane_correct = float2((sign.x+1)*GRASS_SLICE_NUM_INV_DIV2,
                                  (sign.y+1)*GRASS_SLICE_NUM_INV_DIV2);
    float2 planemod = float2(floor(rayEntry.x*PLANE_NUM)/PLANE_NUM,
                             floor(rayEntry.y*PLANE_NUM)/PLANE_NUM);
    float2 pre_dir_correct = float2((sign.x+1)*PLANE_NUM_INV_DIV2,
                                    (sign.y+1)*PLANE_NUM_INV_DIV2);

    int hitcount;
    for(hitcount =0; hitcount < MAX_RAYDEPTH % (MAX_RAYDEPTH+1); hitcount++) // %([MAX_RAYDEPTH]+1) speeds up compilation.
                                         // It may proof to be faster to early exit this loop
                                         // depending on the hardware used.
    {

        //Calculate positions of the intersections with the next grid planes on the u,v tangent space axis independently.

        float2 dir_correct = float2(sign.x*plane_offset.x+pre_dir_correct.x,
                                    sign.y*plane_offset.y+pre_dir_correct.y);           
        float2 distance = float2((planemod.x + dir_correct.x - rayEntry.x)/(i.EyeDirTan.x),
                                 (planemod.y + dir_correct.y - rayEntry.y)/(i.EyeDirTan.y));
                    
        float3 rayHitpointX = rayEntry + i.EyeDirTan *distance.x;   
        float3 rayHitpointY = rayEntry + i.EyeDirTan *distance.y;
        
        //Check if we hit the ground. If so, calculate the intersection and look up the ground texture and blend colors.

        if ((rayHitpointX.z <= -GRASSDEPTH)&& (rayHitpointY.z <= -GRASSDEPTH))  
        {
            float distanceZ = (-GRASSDEPTH)/i.EyeDirTan.z; // rayEntry.z is 0.0 anyway 

            float3 rayHitpointZ = rayEntry + i.EyeDirTan *distanceZ;
            float2 orthoLookupZ = float2(rayHitpointZ.x,rayHitpointZ.y);
            
            //mod: no ground plane
            //color = (color)+((1.0-color.w) * tex2D(ground,orthoLookupZ));
            if(zFlag ==1) zOffset = distanceZ; // write the distance from rayEntry to intersection
            zFlag = 0; //Early exit here if faster.     
        }  
        else
        {
            float2 orthoLookup; //Will contain texture lookup coordinates for grassblades texture.

            //check if we hit a u or v plane, calculate lookup accordingly with wind shear displacement.
            if(distance.x <= distance.y)
            {
                float4 windX = (tex2D(SWindNoise,i.TexCoordWind+rayHitpointX.xy/TC1_TO_TC2_RATIO)-0.5)/2;
                
                float lookupX = -(rayHitpointX.z+(planemod.x+sign.x*plane_offset.x)*PREMULT)-plane_correct.x;
                orthoLookup=float2(rayHitpointX.y+windX.x*(GRASSDEPTH+rayHitpointX.z),lookupX); 
                
                plane_offset.x += PLANE_NUM_INV; // increment/decrement to next grid plane on u axis
                if(zFlag==1) zOffset = distance.x;
            }
            else 
            {
                float4 windY = (tex2D(SWindNoise,i.TexCoordWind+rayHitpointY.xy/TC1_TO_TC2_RATIO)-0.5)/2;
            
                float lookupY = -(rayHitpointY.z+(planemod.y+sign.y*plane_offset.y)*PREMULT)-plane_correct.y;
                orthoLookup = float2(rayHitpointY.x+windY.y*(GRASSDEPTH+rayHitpointY.z) ,lookupY);
            
                plane_offset.y += PLANE_NUM_INV;  // increment/decrement to next grid plane on v axis
                if(zFlag==1) zOffset = distance.y;
                    
            }
        
            color += (1.0-color.w)*tex2D(SBlades, orthoLookup);

            if(color.w >= 0.49){zFlag = 0;} //Early exit here if faster.
        }
    }   

    if(color.w)
        color += (1.0-color.w)*AVERAGE_COLOR; //Fill remaining transparency in case there is some left. Can be replaced by a texture lookup
                       //into a fully opaque grass slice using orthoLookup.
    
   //color = tex2D(SBlades, i.TexCoordGrass);

     //mod: no lighting so far
     //color.xyz *= (vertColor.xyz); //Modulate with per vertex lightmap,as an alternative, modulate with N*L for dynamic lighting.

     //zOffset is along eye direction, transform and add to vertex position to get correct z-value.
    i.PositionViewProj += mul(i.EyeDirTan*zOffset, matVP); 
        
    //Divide by homogenous part.
    o.Depth = i.PositionViewProj.z/i.PositionViewProj.w;
     
    float shadow = 0;
    float s = min(i.Diffuse, 1-shadow);
    color.rgb = (s * LightColor.rgb + AmbientColor.rgb) * color.rgb;

    o.Color = color;
    }
    return o;
}

stateblock RenderStates = stateblock_state
{
    ZEnable = true;
    ZWriteEnable = true;
    ZFunc = LESS;

    Lighting = false;
    Cullmode = CCW;
    AlphaBlendEnable = false;
    AlphaTestEnable = true; 
    AlphaRef = 0;  
    AlphaFunc = GREATER; 
    /*BlendOp = ADD;
    SrcBlend = SRCALPHA;
    DestBlend = INVSRCALPHA;*/
};

technique RenderGrass
{
    pass P0
    {
        StateBlock = (RenderStates);

        VertexShader = compile vs_3_0 BillboardGrass_VP(false);
        PixelShader  = compile ps_3_0 BillboardGrass_PS(false);
    }

}










